import fnmatch
import re
import sys
from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import ClassVar, Dict, List, Optional, Sequence, Set, Tuple, TypeVar, Union

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

if sys.version_info < (3, 10):
    from typing_extensions import Literal, TypeAlias
else:
    from typing import Literal, TypeAlias

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

import pydantic
from packaging.specifiers import SpecifierSet

from sigmatcher.grep import rip_regex
from sigmatcher.results import Result


class InvalidMacroModifierError(Exception):
    """
    Exception raised when an invalid macro modifier is encountered.
    """

    def __init__(self, modifier: str, class_name: str) -> None:
        self.modifier = modifier
        self.class_name = class_name
        super().__init__(f"Invalid macro modifier: '{modifier}' for class '{class_name}'")


def is_in_version_range(app_version: Optional[str], version_range: Union[str, List[str], None]) -> bool:
    if app_version is None or version_range is None:
        return True
    ranges = version_range if isinstance(version_range, list) else [version_range]
    return any(SpecifierSet(spec).contains(app_version) for spec in ranges)


class BaseSignature(ABC, pydantic.BaseModel, frozen=True, use_attribute_docstrings=True):
    version_range: Union[str, List[str], None] = None
    """The version range in which the signature is valid."""
    count: int = 1
    """The number of times the signature should match in order to be considered a match."""

    @abstractmethod
    def check_files(self, search_pathes: List[Path]) -> List[Path]:
        raise NotImplementedError()

    @abstractmethod
    def check_strings(self, strings: List[str]) -> List[str]:
        raise NotImplementedError()

    @abstractmethod
    def capture(self, value: str) -> Set[str]:
        raise NotImplementedError()

    @abstractmethod
    def get_dependencies(self) -> List[str]:
        raise NotImplementedError()

    def resolve_macro(
        self, results: Dict[str, Union[Result, Exception]], result_identifier: str, result_modifier: str
    ) -> str:
        result = results[result_identifier]
        assert not isinstance(result, Exception)

        try:
            resolved_macro = getattr(result.new, result_modifier)
        except AttributeError:
            raise InvalidMacroModifierError(result_modifier, result.new.__class__.__name__) from None

        assert isinstance(resolved_macro, str)
        return resolved_macro

    @abstractmethod
    def resolve_macros(self, results: Dict[str, Union[Result, Exception]]) -> Self:
        raise NotImplementedError()

    def is_in_version_range(self, app_version: Optional[str]) -> bool:
        return is_in_version_range(app_version, self.version_range)


class BaseRegexSignature(BaseSignature, pydantic.BaseModel, frozen=True):
    signature: "re.Pattern[str]" = pydantic.Field(
        json_schema_extra={"x-intellij-language-injection": {"language": "RegExp"}}
    )
    """
    A regular expression used to check the signature.
    
    When used for capturing strings, if there is more than one match,
    a named group called `match` should be used to specify what should be captured.
    
    Can include macros in the form of `${<result>.<property>}`.
    Results are actually objects from `sigmatcher.results` package.
    i.e. `Class`, `Field`, `Method`, `Export` objects.
    Property could be any python property that the objects holds.

    For example a macro could look like `${MyClass.fields.java}`, which would return something like
    `Lf9/s;->a:LX/Y/Z/A02;`
    """

    MACRO_REGEX: "ClassVar[re.Pattern[str]]" = re.compile(r"\${(.*?)}")

    def check_files(self, search_pathes: List[Path]) -> List[Path]:
        return [
            path.resolve()
            for path, match_count in rip_regex(self.signature, search_pathes).items()
            if self.count in (match_count, 0)
        ]

    def check_strings(self, strings: List[str]) -> List[str]:
        results: List[str] = []
        for string in strings:
            match_count = len(self.signature.findall(string))
            if match_count == 0:
                continue
            if self.count in (match_count, 0):
                results.append(string)
        return results

    def capture(self, value: str) -> Set[str]:
        match = self.signature.search(value)
        if match is None:
            return set()
        try:
            return {match.group("match")}
        except IndexError:
            pass

        return set(match.groups())

    def get_dependencies(self) -> List[str]:
        return [macro.rpartition(".")[0] for macro in self._get_raw_macros]

    @cached_property
    def _get_raw_macros(self) -> Set[str]:
        return set(self.MACRO_REGEX.findall(self.signature.pattern))

    def resolve_macros(self, results: Dict[str, Union[Result, Exception]]) -> Self:
        if not self._get_raw_macros:
            return self

        new_pattern = self.signature.pattern
        for macro in self._get_raw_macros:
            result_identifier, _, result_modifier = macro.rpartition(".")
            resolved_macro = self.resolve_macro(results, result_identifier, result_modifier)
            new_pattern = new_pattern.replace(f"${{{macro}}}", re.escape(resolved_macro))

        return self.model_copy(update={"signature": re.compile(new_pattern)})


class RegexSignature(BaseRegexSignature, frozen=True):
    type: Literal["regex"] = "regex"
    """The type of the signature."""


class GlobSignature(BaseRegexSignature, frozen=True):
    type: Literal["glob"] = "glob"
    """The type of the signature."""

    @pydantic.field_validator("signature", mode="before")
    @classmethod
    def parse_glob(cls, v: str) -> str:
        # TODO: removing the atomic group, i.e. the "(?>" makes glob signature in the form of "*WORD*" slower then
        #  their regex counterparts
        return fnmatch.translate(v).replace("\\Z", "$").replace("(?>", "(?:")


class TreeSitterSignature(BaseSignature, pydantic.BaseModel, frozen=True):
    signature: str
    """A TreeSitter s-query used to check the signature."""
    type: Literal["treesitter"] = "treesitter"
    """The type of the signature."""

    def check_files(self, search_pathes: List[Path]) -> List[Path]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    def check_strings(self, strings: List[str]) -> List[str]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    def capture(self, value: str) -> Set[str]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    def get_dependencies(self) -> List[str]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    def resolve_macros(self, results: Dict[str, Union[Result, Exception]]) -> Self:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")


Signature: TypeAlias = Annotated[
    Union[RegexSignature, GlobSignature, TreeSitterSignature], pydantic.Field(discriminator="type")
]


class Definition(pydantic.BaseModel, frozen=True, use_attribute_docstrings=True):
    name: str
    """The name of the definition, i.e. the class, method, field, or export name."""
    signatures: Tuple[Signature, ...]
    """A list of signatures that define the definition."""
    version_range: Union[str, List[str], None] = None
    """The version range in which the definition is valid."""

    def get_signatures_for_version(self, app_version: Optional[str]) -> Tuple[Signature, ...]:
        if app_version is None:
            return self.signatures
        return tuple(signature for signature in self.signatures if signature.is_in_version_range(app_version))

    def get_dependencies(self, app_version: Optional[str]) -> Set[str]:
        dependencies: Set[str] = set()
        for signature in self.get_signatures_for_version(app_version):
            dependencies.update(signature.get_dependencies())
        return dependencies

    def is_in_version_range(self, app_version: Optional[str]) -> bool:
        return is_in_version_range(app_version, self.version_range)


class ExportDefinition(Definition, frozen=True):
    pass


class FieldDefinition(Definition, frozen=True):
    pass


class MethodDefinition(Definition, frozen=True):
    pass


class ClassDefinition(Definition, frozen=True):
    package: Optional[str] = None
    """The package of the class."""
    fields: Tuple[FieldDefinition, ...] = ()
    """A list of field definitions."""
    methods: Tuple[MethodDefinition, ...] = ()
    """A list of method definitions."""
    exports: Tuple[ExportDefinition, ...] = ()
    """A list of export definitions."""


DEFINITIONS_TYPE_ADAPTER = pydantic.TypeAdapter(List[ClassDefinition])


TDefinition = TypeVar("TDefinition", bound=Definition)


def merge_definition(def1: TDefinition, def2: TDefinition) -> TDefinition:
    signatures = def2.signatures if def2.signatures else def1.signatures

    if isinstance(def1, ClassDefinition):
        assert isinstance(def2, ClassDefinition)
        fields_to_update = {
            "signatures": signatures,
            "methods": merge_definitions_groups([def1.methods, def2.methods]),
            "fields": merge_definitions_groups([def1.fields, def2.fields]),
            "exports": merge_definitions_groups([def1.exports, def2.exports]),
        }
    else:
        fields_to_update = {"signatures": signatures}

    return def1.model_copy(update=fields_to_update)


def merge_definitions_groups(definition_groups: Sequence[Sequence[TDefinition]]) -> Tuple[TDefinition, ...]:
    """
    Merge multiple definition groups into a single group,
    e.g. merge all the methods and fields of 2 class definitions with the same name.
    When a definition is present in multiple groups, the signatures are taken from the last group.
    """
    final_definitions: Dict[str, TDefinition] = {}
    for definition_group in definition_groups:
        for definition in definition_group:
            if definition.name in final_definitions:
                definition = merge_definition(final_definitions[definition.name], definition)
            final_definitions[definition.name] = definition
    return tuple(final_definitions.values())



import hashlib
import json
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

import platformdirs
import pydantic
import pydantic_core
import typer
import yaml
from rich.console import Console

import sigmatcher.analysis
from sigmatcher import __version__
from sigmatcher.definitions import DEFINITIONS_TYPE_ADAPTER, ClassDefinition, merge_definitions_groups
from sigmatcher.formats import OutputFormat, convert_to_format
from sigmatcher.results import MatchedClass

app = typer.Typer()

stdout_console = Console(soft_wrap=True)
stderr_console = Console(soft_wrap=True, stderr=True)

cache_app = typer.Typer(help="Manage Sigmatcher's cache")
app.add_typer(cache_app, name="cache")

CACHE_DIR_PATH = platformdirs.user_cache_path("sigmatcher", "oriori1703", ensure_exists=True)


@cache_app.command()
def info() -> None:
    """
    Get the path to Sigmatcher's cache directory.
    """
    print(str(CACHE_DIR_PATH))


@cache_app.command()
def clean() -> None:
    """
    Clean the cache directory.
    """
    for path in CACHE_DIR_PATH.iterdir():
        shutil.rmtree(path)
    stdout_console.print("[green]Successfully cleaned the cache directory.[/green]")


def version_callback(value: bool) -> None:
    if value:
        stdout_console.print(f"Sigmatcher version: [green]{__version__}[/green]")
        raise typer.Exit()


@app.callback()
def callback(
    _version: Annotated[
        Optional[bool],
        typer.Option("--version", help="Show the version and exit.", callback=version_callback),
    ] = None,
) -> None:
    pass


@app.command()
def schema(
    output: Annotated[
        Optional[Path],
        typer.Option(help="Output path for the json schema. If none is given print to stdout"),
    ] = None,
) -> None:
    """
    Get the json schema for writing definitions.
    """

    class SigmatcherGenerateJsonSchema(pydantic.json_schema.GenerateJsonSchema):
        def generate(
            self, schema: pydantic_core.CoreSchema, mode: pydantic.json_schema.JsonSchemaMode = "validation"
        ) -> pydantic.json_schema.JsonSchemaValue:
            json_schema = super().generate(schema, mode=mode)
            json_schema["title"] = "Sigmatcher's Definitions"
            json_schema["$schema"] = self.schema_dialect
            return json_schema

    definitions_schema = DEFINITIONS_TYPE_ADAPTER.json_schema(schema_generator=SigmatcherGenerateJsonSchema)
    definitions_schema_json = json.dumps(definitions_schema, indent=2)
    if output is not None:
        output.write_text(definitions_schema_json)
    else:
        stdout_console.print(definitions_schema_json)


def apktool_callback(value: str) -> str:
    if shutil.which(value) is None:
        raise typer.BadParameter("Cannot find the apktool executable")
    return value


@app.command()
def analyze(
    apk: Annotated[
        Path, typer.Argument(help="Path to the apk that will be analyzed", exists=True, file_okay=True, dir_okay=False)
    ],
    signatures: Annotated[
        List[Path],
        typer.Option(
            help="Path to a signature file. If multiple files are given, they are merged together",
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
    ],
    output_file: Annotated[
        Optional[Path],
        typer.Option(help="Output path for the final mapping output"),
    ] = None,
    output_format: Annotated[OutputFormat, typer.Option(help="The output mapping format")] = OutputFormat.RAW,
    apktool: Annotated[
        str, typer.Option(help="The command to use when running apktool", callback=apktool_callback)
    ] = "apktool",
) -> None:
    definition_groups: List[Tuple[ClassDefinition, ...]] = []
    for signature_file in signatures:
        with signature_file.open("r") as f:
            raw_yaml = yaml.safe_load(f)
        definitions = DEFINITIONS_TYPE_ADAPTER.validate_python(raw_yaml)
        definition_groups.append(tuple(definitions))
    merged_definitions = merge_definitions_groups(definition_groups)

    apk_hash = hashlib.sha256(apk.read_bytes()).hexdigest()
    unpacked_path = CACHE_DIR_PATH / apk_hash
    if not unpacked_path.exists():
        subprocess.run([apktool, "decode", apk, "--output", unpacked_path])

    apktool_yaml_file = unpacked_path / "apktool.yml"
    with apktool_yaml_file.open() as f:
        apk_version = yaml.safe_load(f)["versionInfo"]["versionName"]
    assert isinstance(apk_version, str)

    results = sigmatcher.analysis.analyze(merged_definitions, unpacked_path, apk_version)
    successful_results: Dict[str, MatchedClass] = {}
    for analyzer_name, result in results.items():
        if isinstance(result, Exception):
            stderr_console.print(f"[yellow]Error in {analyzer_name} - {result!s}[/yellow]")
            continue
        if isinstance(result, MatchedClass):
            successful_results[analyzer_name] = result

    mapping_output = convert_to_format(successful_results, output_format)
    if output_file is None:
        stdout_console.print(mapping_output)
    else:
        output_file.write_text(mapping_output)


def main() -> None:
    app()



import enum
import json
from abc import ABC, abstractmethod
from io import StringIO
from typing import Dict, Type

import pydantic

from sigmatcher.results import MatchedClass, MatchedField, MatchedMethod


class Formatter(ABC):
    @abstractmethod
    def convert(self, matched_classes: Dict[str, MatchedClass]) -> str:
        raise NotImplementedError()


class RawFormatter(Formatter):
    def convert(self, matched_classes: Dict[str, MatchedClass]) -> str:
        return pydantic.RootModel[Dict[str, MatchedClass]](matched_classes).model_dump_json(indent=4)


class LegacyFormatter(Formatter):
    def convert(self, matched_classes: Dict[str, MatchedClass]) -> str:
        return json.dumps(
            {
                matched_class.original.name: {
                    "className": f"{matched_class.new.package}.{matched_class.new.name}",
                    "methods": {method.original.name: method.new.name for method in matched_class.matched_methods},
                    "fields": {field.original.name: field.new.name for field in matched_class.matched_fields},
                }
                for matched_class in matched_classes.values()
            },
            indent=4,
            sort_keys=True,
        )


class EnigmaFormatter(Formatter):
    def convert_field(self, field: MatchedField) -> str:
        return f"\tFIELD {field.new.name} {field.original.name} {field.new.type}\n"

    def convert_method(self, method: MatchedMethod) -> str:
        return (
            f"\tMETHOD {method.new.name} {method.original.name} ({method.new.argument_types}){method.new.return_type}\n"
        )

    def convert_class(self, matched_class: MatchedClass) -> str:
        result = StringIO()
        new_class = matched_class.new.to_java_representation()[1:-1]
        original_class = matched_class.original.to_java_representation()[1:-1]
        result.write(f"CLASS {new_class} {original_class}\n")
        for field in matched_class.matched_fields:
            result.write(self.convert_field(field))
        for method in matched_class.matched_methods:
            result.write(self.convert_method(method))
        return result.getvalue()

    def convert(self, matched_classes: Dict[str, MatchedClass]) -> str:
        final = StringIO()
        for matched_class in matched_classes.values():
            final.write(self.convert_class(matched_class))
        return final.getvalue()


class OutputFormat(str, enum.Enum):
    RAW = "raw"
    ENIGMA = "enigma"
    LEGACY = "legacy"


FORMAT_TO_FORMATTER: Dict[OutputFormat, Type[Formatter]] = {
    OutputFormat.RAW: RawFormatter,
    OutputFormat.ENIGMA: EnigmaFormatter,
    OutputFormat.LEGACY: LegacyFormatter,
}


def convert_to_format(matched_classes: Dict[str, MatchedClass], output_format: OutputFormat) -> str:
    return FORMAT_TO_FORMATTER[output_format]().convert(matched_classes)



import re
import subprocess
from pathlib import Path
from typing import Dict, List, Tuple, Union


def parse_rg_line(line: str) -> Tuple[Path, int]:
    path, _, count = line.rpartition(":")
    return Path(path), int(count)


def rip_regex(pattern: "Union[str, re.Pattern[str]]", search_paths: List[Path]) -> Dict[Path, int]:
    if isinstance(pattern, re.Pattern):
        pattern = pattern.pattern
    args = [
        "rg",
        "--count-matches",
        "--multiline",
        "--no-ignore",
        "--hidden",
        "--regexp",
        pattern,
        *[str(path) for path in search_paths],
    ]
    process = subprocess.run(args, stdout=subprocess.PIPE, text=True)

    if bool(process.returncode):
        return {}

    if len(search_paths) == 1 and search_paths[0].is_file():
        # if the search path is a single file, ripgrep will not print the filename
        return {search_paths[0]: int(process.stdout)}

    return dict(parse_rg_line(line) for line in process.stdout.splitlines())



import dataclasses
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Sequence, Set, Tuple, TypeVar, Union

import graphlib

from sigmatcher.definitions import (
    ClassDefinition,
    Definition,
    ExportDefinition,
    FieldDefinition,
    InvalidMacroModifierError,
    MethodDefinition,
    Signature,
)
from sigmatcher.results import (
    Class,
    Field,
    MatchedClass,
    MatchedExport,
    MatchedField,
    MatchedMethod,
    Method,
    Result,
)


class MatchError(Exception):
    pass


class NoMatchesError(MatchError):
    pass


class TooManyMatchesError(MatchError):
    pass


class DependencyMatchError(MatchError):
    pass


SignatureMatch = TypeVar("SignatureMatch", str, Path)


def filter_signature_matches(
    matches_per_signature: Iterable[Tuple[Signature, List[SignatureMatch]]],
) -> Set[SignatureMatch]:
    whitelist_matches: Set[SignatureMatch] = set()
    blacklist_matches: Set[SignatureMatch] = set()
    first_match = True

    for signature, matches in matches_per_signature:
        if signature.count == 0:
            blacklist_matches.update(matches)
        elif first_match:
            first_match = False
            whitelist_matches.update(matches)
        else:
            whitelist_matches.intersection_update(matches)
            if not len(whitelist_matches):
                return whitelist_matches
    whitelist_matches.difference_update(blacklist_matches)
    return whitelist_matches


@dataclasses.dataclass(frozen=True)
class Analyzer(ABC):
    definition: Definition
    app_version: Optional[str]

    @abstractmethod
    def analyze(self, results: Dict[str, Union[Result, Exception]]) -> Result:
        pass

    def check_match_count(self, matches: Optional[Set[SignatureMatch]]) -> None:
        if matches is None or len(matches) == 0:
            raise NoMatchesError(f"Found no match for {self.name}!")
        if len(matches) > 1:
            raise TooManyMatchesError(f"Found too many matches for {self.name}: {matches}")

    def get_dependencies(self) -> Set[str]:
        return self.definition.get_dependencies(self.app_version)

    def check_dependencies(self, results: Dict[str, Union[Result, Exception]]) -> None:
        failed_dependencies: List[str] = []
        for dependency_name in self.get_dependencies():
            child_result = results[dependency_name]
            if isinstance(child_result, Exception):
                failed_dependencies.append(dependency_name)

        if failed_dependencies:
            raise DependencyMatchError(
                f"Skipped {self.name} because of the following dependencies failed: {failed_dependencies}"
            )

    def get_signatures_for_version(self) -> Tuple[Signature, ...]:
        return self.definition.get_signatures_for_version(self.app_version)

    @property
    def name(self) -> str:
        return self.definition.name

    def __repr__(self) -> str:
        return self.name


@dataclasses.dataclass(frozen=True)
class ClassAnalyzer(Analyzer):
    definition: ClassDefinition
    search_root: Path

    def analyze(self, results: Dict[str, Union[Result, Exception]]) -> MatchedClass:
        signatures = list(self.get_signatures_for_version())
        if len(signatures) == 0:
            raise NoMatchesError(f"Found no signatures for {self.name}! Make sure your version ranges are correct.")

        # Make sure the first signature is a whitelist signature in order to improve performance
        whitelist_signature_index = 0
        for i, signature in enumerate(signatures):
            if signature.count != 0:
                whitelist_signature_index = i
                break
        signatures.insert(0, signatures.pop(whitelist_signature_index))

        class_matches = set(self.search_root.rglob("*.smali"))

        for signature in signatures:
            # Limit the search avoid too many arguments to ripgrep
            if len(class_matches) < 100:
                search_paths = list(class_matches)
            else:
                search_paths = [self.search_root]

            # TODO: Dedeuplicate this and filter_signature_matches()
            matches = signature.resolve_macros(results).check_files(search_paths)
            if signature.count == 0:
                class_matches.difference_update(matches)
            else:
                class_matches.intersection_update(matches)
                if len(class_matches) == 0:
                    break

        self.check_match_count(class_matches)
        match = next(iter(class_matches))
        with match.open() as f:
            class_definition_line = f.readline().rstrip("\n")
        _, _, raw_class_name = class_definition_line.rpartition(" ")
        new_class = Class.from_java_representation(raw_class_name)
        original_class = Class(name=self.definition.name, package=self.definition.package or new_class.package)
        return MatchedClass(
            original=original_class, new=new_class, smali_file=match, matched_methods=[], matched_fields=[]
        )


@dataclasses.dataclass(frozen=True)
class FieldAnalyzer(Analyzer):
    definition: FieldDefinition
    parent: ClassAnalyzer

    def analyze(self, results: Dict[str, Union[Result, Exception]]) -> MatchedField:
        parent_class_result = results[self.parent.name]
        assert isinstance(parent_class_result, MatchedClass)

        raw_class = parent_class_result.smali_file.read_text()
        signatures = self.get_signatures_for_version()
        if len(signatures) == 0:
            raise NoMatchesError(f"Found no signatures for {self.name}! Make sure your version ranges are correct.")
        if len(signatures) > 1:
            raise NoMatchesError(
                f"Found {len(signatures)} signatures for {self.name}. Field definitions should only have one."
            )
        signature = signatures[0].resolve_macros(results)
        captured_names = signature.capture(raw_class)
        self.check_match_count(captured_names)
        raw_field_name = next(iter(captured_names))

        new_field = Field.from_java_representation(raw_field_name)
        # TODO: should we get the types for the original field from the definition?
        original_field = Field(name=self.definition.name, type=new_field.type)
        matched_field = MatchedField(original=original_field, new=new_field)
        parent_class_result.matched_fields.append(matched_field)
        return matched_field

    def get_dependencies(self) -> Set[str]:
        return super().get_dependencies() | {self.parent.name}

    @property
    def name(self) -> str:
        return f"{self.parent.name}.fields.{self.definition.name}"


@dataclasses.dataclass(frozen=True)
class MethodAnalyzer(Analyzer):
    definition: MethodDefinition
    parent: ClassAnalyzer

    def analyze(self, results: Dict[str, Union[Result, Exception]]) -> MatchedMethod:
        parent_class_result = results[self.parent.name]
        assert isinstance(parent_class_result, MatchedClass)

        raw_methods = parent_class_result.smali_file.read_text().split(".method")[1:]
        methods = [".method" + method for method in raw_methods]

        signatures = self.get_signatures_for_version()
        if len(signatures) == 0:
            raise NoMatchesError(f"Found no signatures for {self.name}! Make sure your version ranges are correct.")

        method_matches = filter_signature_matches(
            (signature, signature.resolve_macros(results).check_strings(methods)) for signature in signatures
        )
        self.check_match_count(method_matches)
        match = next(iter(method_matches))
        method_definition_line, _, _ = match.partition("\n")
        _, _, raw_method_name = method_definition_line.rpartition(" ")

        new_method = Method.from_java_representation(raw_method_name)
        # TODO: should we get the types for the original method from the definition?
        original_method = Method(
            name=self.definition.name, argument_types=new_method.argument_types, return_type=new_method.return_type
        )
        matched_method = MatchedMethod(original=original_method, new=new_method)
        parent_class_result.matched_methods.append(matched_method)
        return matched_method

    def get_dependencies(self) -> Set[str]:
        return super().get_dependencies() | {self.parent.name}

    @property
    def name(self) -> str:
        return f"{self.parent.name}.methods.{self.definition.name}"


@dataclasses.dataclass(frozen=True)
class ExportAnalyzer(Analyzer):
    definition: ExportDefinition
    parent: ClassAnalyzer

    def analyze(self, results: Dict[str, Union[Result, Exception]]) -> MatchedExport:
        parent_class_result = results[self.parent.name]
        assert isinstance(parent_class_result, MatchedClass)

        raw_class = parent_class_result.smali_file.read_text()
        signatures = self.get_signatures_for_version()
        signature = signatures[0].resolve_macros(results)
        if len(signatures) == 0:
            raise NoMatchesError(f"Found no signatures for {self.name}! Make sure your version ranges are correct.")
        if len(signatures) > 1:
            raise NoMatchesError(
                f"Found {len(signatures)} signatures for {self.name}. Export definitions should only have one."
            )

        captured_names = signature.capture(raw_class)
        self.check_match_count(captured_names)
        export_value = next(iter(captured_names))

        return MatchedExport.from_value(export_value)

    def get_dependencies(self) -> Set[str]:
        return super().get_dependencies() | {self.parent.name}

    @property
    def name(self) -> str:
        return f"{self.parent.name}.exports.{self.definition.name}"


def create_analyzers(
    definitions: Sequence[ClassDefinition], search_root: Path, app_version: Optional[str]
) -> Dict[str, Analyzer]:
    canonical_search_root = search_root.resolve()
    name_to_analyzer: Dict[str, Analyzer] = {}
    for class_definition in definitions:
        if not class_definition.is_in_version_range(app_version):
            continue

        class_analyzer = ClassAnalyzer(class_definition, app_version, canonical_search_root)
        name_to_analyzer[class_definition.name] = class_analyzer

        for method_definition in class_definition.methods:
            if not method_definition.is_in_version_range(app_version):
                continue
            method_analyzer = MethodAnalyzer(method_definition, app_version, parent=class_analyzer)
            name_to_analyzer[method_analyzer.name] = method_analyzer

        for field_definition in class_definition.fields:
            if not field_definition.is_in_version_range(app_version):
                continue
            field_analyzer = FieldAnalyzer(field_definition, app_version, parent=class_analyzer)
            name_to_analyzer[field_analyzer.name] = field_analyzer

        for export_definition in class_definition.exports:
            if not export_definition.is_in_version_range(app_version):
                continue
            export_analyzer = ExportAnalyzer(export_definition, app_version, parent=class_analyzer)
            name_to_analyzer[export_analyzer.name] = export_analyzer

    return name_to_analyzer


def analyze(
    definitions: Sequence[ClassDefinition], unpacked_path: Path, app_version: Optional[str]
) -> Dict[str, Union[Result, Exception]]:
    name_to_analyzer = create_analyzers(definitions, unpacked_path, app_version)

    sorter: graphlib.TopologicalSorter[str] = graphlib.TopologicalSorter()
    for analyzer in name_to_analyzer.values():
        sorter.add(analyzer.name, *analyzer.get_dependencies())

    results: Dict[str, Union[Result, Exception]] = {}
    for analyzer_name in sorter.static_order():
        analyzer = name_to_analyzer[analyzer_name]
        try:
            analyzer.check_dependencies(results)
            results[analyzer_name] = analyzer.analyze(results)
        except (MatchError, InvalidMacroModifierError) as e:
            results[analyzer_name] = e
    return results
